import json
import urllib
from urllib.parse import urlparse

from django.utils.translation import gettext as _
from fhir.resources.address import Address

from insuree.models import Insuree, Gender, Education, Profession, Family, \
    InsureePhoto, Relation, IdentificationType
from location.models import Location, HealthFacility
from api_fhir_r4.configurations import R4IdentifierConfig, GeneralConfiguration, R4MaritalConfig
from api_fhir_r4.converters import BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin
from api_fhir_r4.converters.groupConverter import GroupConverter
from api_fhir_r4.converters.locationConverter import LocationConverter
from api_fhir_r4.mapping.patientMapping import RelationshipMapping, EducationLevelMapping, \
    PatientProfessionMapping, MaritalStatusMapping, PatientCategoryMapping
from api_fhir_r4.models.imisModelEnums import ImisMaritalStatus
from fhir.resources.patient import Patient, PatientContact
from fhir.resources.extension import Extension
from fhir.resources.attachment import Attachment
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.utils import TimeUtils, DbManagerUtils


class PatientConverter(BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_insuree, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_patient = Patient.construct()
        cls.build_fhir_pk(fhir_patient, imis_insuree, reference_type)
        cls.build_human_names(fhir_patient, imis_insuree)
        cls.build_fhir_identifiers(fhir_patient, imis_insuree)
        cls.build_fhir_birth_date(fhir_patient, imis_insuree)
        cls.build_fhir_gender(fhir_patient, imis_insuree)
        cls.build_fhir_marital_status(fhir_patient, imis_insuree)
        cls.build_fhir_telecom(fhir_patient, imis_insuree)
        cls.build_fhir_addresses(fhir_patient, imis_insuree, reference_type)
        cls.build_fhir_extentions(fhir_patient, imis_insuree, reference_type)
        cls.build_fhir_contact(fhir_patient, imis_insuree)
        cls.build_fhir_photo(fhir_patient, imis_insuree)
        cls.build_fhir_general_practitioner(fhir_patient, imis_insuree, reference_type)
        return fhir_patient

    @classmethod
    def to_imis_obj(cls, fhir_patient, audit_user_id):
        errors = []
        fhir_patient = Patient(**fhir_patient)
        imis_insuree = cls.createDefaultInsuree(audit_user_id)
        cls.build_imis_names(imis_insuree, fhir_patient, errors)
        cls.build_imis_identifiers(imis_insuree, fhir_patient)
        cls.build_imis_birth_date(imis_insuree, fhir_patient, errors)
        cls.build_imis_gender(imis_insuree, fhir_patient)
        cls.build_imis_marital(imis_insuree, fhir_patient)
        cls.build_imis_contacts(imis_insuree, fhir_patient)
        cls.build_imis_addresses(imis_insuree, fhir_patient)
        cls.build_imis_photo(imis_insuree, fhir_patient, errors)
        cls.build_imis_extentions(imis_insuree, fhir_patient, errors)
        cls.build_imis_family(imis_insuree, fhir_patient, errors)
        cls.build_imis_relationship(imis_insuree, fhir_patient)
        cls.build_imis_general_practitioner(imis_insuree, fhir_patient)
        return imis_insuree

    @classmethod
    def build_fhir_pk(cls, fhir_patient, resource, reference_type: str = None):
        if reference_type == ReferenceConverterMixin.CODE_REFERENCE_TYPE:
            fhir_patient.id = resource.chf_id
        else:
            super().build_fhir_pk(fhir_patient, resource, reference_type)

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def get_reference_obj_uuid(cls, imis_patient: Insuree):
        return imis_patient.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_patient: Insuree):
        return imis_patient.id

    @classmethod
    def get_reference_obj_code(cls, imis_patient: Insuree):
        return imis_patient.chf_id

    @classmethod
    def build_imis_extentions(cls, imis_insuree, fhir_patient, errors):
        cls._validate_fhir_extension_is_exist(fhir_patient)
        for extension in fhir_patient.extension:
            if extension.url == f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-is-head":
                imis_insuree.head = extension.valueBoolean

            elif extension.url == f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-education-level":
                try:
                    imis_insuree.education = Education.objects.get(id=extension.valueCodeableConcept.coding[0].code)
                except:
                    imis_insuree.education = None

            elif extension.url == f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-profession":
                try:
                    imis_insuree.profession = Profession.objects.get(id=extension.valueCodeableConcept.coding[0].code)
                except:
                    imis_insuree.profession = None

            elif extension.url == f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-card-issued":
                try:
                    imis_insuree.card_issued = extension.valueBoolean
                except:
                    imis_insuree.card_issued = False

            elif extension.url == f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-identification":
                try:
                    for ext in extension.extension:
                        if ext.url == "number":
                            imis_insuree.passport = ext.valueString
                        if ext.url == "type":
                            imis_insuree.type_of_id = IdentificationType.objects.get(code=ext.valueCodeableConcept.coding[0].code)
                except:
                    imis_insuree.passport = None
                    imis_insuree.type_of_id = None
            else:
                pass
        cls._validate_imis_is_head(imis_insuree)
    
    @classmethod
    def get_location_reference(cls, location):
        return location.rsplit('Location/', 1)[1]

    @classmethod
    def get_fhir_resource_type(cls):
        return Patient

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_insuree_uuid = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Insuree, uuid=imis_insuree_uuid)

    @classmethod
    def createDefaultInsuree(cls, audit_user_id):
        imis_insuree = Insuree()
        # temporary set uuid as None - this will be generated in service create insuree from that module
        imis_insuree.uuid = None
        imis_insuree.head = GeneralConfiguration.get_default_value_of_patient_head_attribute()
        imis_insuree.card_issued = GeneralConfiguration.get_default_value_of_patient_card_issued_attribute()
        imis_insuree.validity_from = TimeUtils.now()
        imis_insuree.audit_user_id = audit_user_id
        return imis_insuree

    @classmethod
    def build_human_names(cls, fhir_patient, imis_insuree):
        name = cls.build_fhir_names_for_person(imis_insuree)
        if type(fhir_patient.name) is not list:
            fhir_patient.name = [name]
        else:
            fhir_patient.name.append(name)

    @classmethod
    def build_imis_names(cls, imis_insuree, fhir_patient, errors):
        cls._validate_fhir_patient_human_name(fhir_patient)
        names = fhir_patient.name
        imis_insuree.last_name, imis_insuree.other_names = cls.build_imis_last_and_other_name(names)
        cls._validate_imis_insuree_human_name(imis_insuree)

    @classmethod
    def build_fhir_identifiers(cls, fhir_patient, imis_insuree):
        identifiers = []
        cls._validate_imis_identifier_code(imis_insuree)
        cls.build_all_identifiers(identifiers, imis_insuree)
        cls.build_fhir_passport_identifier(identifiers, imis_insuree)
        fhir_patient.identifier = identifiers
        cls._validate_fhir_identifier_is_exist(fhir_patient)

    @classmethod
    def build_fhir_code_identifier(cls, identifiers, imis_object: Insuree):
        # Patient don't have code so chfid is used instead as code identifier
        if hasattr(imis_object, 'chf_id'):
            identifiers.append(cls.__build_chfid_identifier(imis_object.chf_id))

    @classmethod
    def __build_chfid_identifier(cls, chfid):
        return cls.build_fhir_identifier(chfid,
                                         R4IdentifierConfig.get_fhir_identifier_type_system(),
                                         R4IdentifierConfig.get_fhir_generic_type_code())

    @classmethod
    def build_imis_identifiers(cls, imis_insuree, fhir_patient):
        insuree_ids = fhir_patient.identifier

        if chf_id := cls.get_fhir_identifier_by_code(insuree_ids, R4IdentifierConfig.get_fhir_generic_type_code()):
            imis_insuree.chf_id = chf_id
        else:
            raise FHIRException("Patient code not provided.")

        if passport := cls.get_fhir_identifier_by_code(insuree_ids, R4IdentifierConfig.get_fhir_passport_type_code()):
            imis_insuree.passport = passport

    @classmethod
    def build_fhir_chfid_identifier(cls, identifiers, imis_insuree):
        cls._validate_imis_identifier_code(imis_insuree)
        identifier = cls.build_fhir_identifier(
            imis_insuree.chf_id,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )
        identifiers.append(identifier)

    @classmethod
    def build_fhir_passport_identifier(cls, identifiers, imis_insuree):
        if hasattr(imis_insuree, "type_of_id") and imis_insuree.type_of_id is not None:
            pass  # TODO typeofid isn't provided, this section should contain logic used to create passport field based on typeofid
        elif imis_insuree.passport:
            identifier = cls.build_fhir_identifier(imis_insuree.passport,
                                                   R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                   R4IdentifierConfig.get_fhir_passport_type_code())
            identifiers.append(identifier)

    @classmethod
    def build_fhir_birth_date(cls, fhir_patient, imis_insuree):
        from core import datetime
        # check if datetime object
        if isinstance(imis_insuree.dob, datetime.datetime):
            fhir_patient.birthDate = str(imis_insuree.dob.date().isoformat())
        else:
            fhir_patient.birthDate = str(imis_insuree.dob.isoformat())
        
    @classmethod
    def build_imis_family(cls, imis_insuree, fhir_patient, errors):
        # Get UUID from patient group reference related extension
        family = cls.__get_family_from_fhir_patient_extension(fhir_patient)
        if family:
            imis_insuree.family = family

        if family and imis_insuree.head:
            raise FHIRException("Patient assigned to existing family can't be head")

    @classmethod
    def build_imis_link(cls, imis_insuree,fhir_link):
        patient = fhir_link[0].other.reference
        value = patient.rsplit('/',1)[1]
        return value
    
    @classmethod
    def build_imis_relationship(cls, imis_insuree,fhir_patient):
        if fhir_patient.contact:
            for contact in fhir_patient.contact:
                if contact.relationship:
                    relationship_name = None
                    for relationship in contact.relationship:
                        for coding in relationship.coding:
                            if "CodeSystem/patient-contact-relationship" in coding.system:
                                relationship_name = coding.display
                    try:
                        relation = Relation.objects.get(relation=relationship_name)
                        imis_insuree.relationship = relation
                    except:
                        pass
    
    @classmethod
    def build_imis_birth_date(cls, imis_insuree, fhir_patient, errors):
        birth_date = fhir_patient.birthDate
        if not cls.valid_condition(birth_date is None, _('Missing patient `birthDate` attribute'), errors):
            imis_insuree.dob = TimeUtils.str_to_date(birth_date)

    @classmethod
    def build_fhir_gender(cls, fhir_patient, imis_insuree):
        if imis_insuree.gender:
            code = imis_insuree.gender.code
            if code == GeneralConfiguration.get_male_gender_code():
                fhir_patient.gender = "male"
            elif code == GeneralConfiguration.get_female_gender_code():
                fhir_patient.gender = "female"
            elif code == GeneralConfiguration.get_other_gender_code():
                fhir_patient.gender = "other"
        else:
            fhir_patient.gender = "unknown"

    @classmethod
    def build_imis_gender(cls, imis_insuree, fhir_patient):
        gender = fhir_patient.gender
        if gender is not None:
            imis_gender = PatientCategoryMapping.imis_gender_mapping.get(gender)
            imis_insuree.gender = imis_gender

    @classmethod
    def build_fhir_marital_status(cls, fhir_patient, imis_insuree):
        if imis_insuree.marital:
            display = MaritalStatusMapping.marital_status[imis_insuree.marital]
            fhir_patient.maritalStatus = \
                cls.build_codeable_concept(code=imis_insuree.marital,
                                           system=R4MaritalConfig.get_fhir_marital_status_system())
            if len(fhir_patient.maritalStatus.coding) == 1:
                fhir_patient.maritalStatus.coding[0].display = display

    @classmethod
    def build_imis_marital(cls, imis_insuree, fhir_patient):
        marital_status = fhir_patient.maritalStatus
        if marital_status is not None:
            for maritialCoding in marital_status.coding:
                if maritialCoding.system == R4MaritalConfig.get_fhir_marital_status_system():
                    code = maritialCoding.code
                    if code == R4MaritalConfig.get_fhir_married_code():
                        imis_insuree.marital = ImisMaritalStatus.MARRIED.value
                    elif code == R4MaritalConfig.get_fhir_never_married_code():
                        imis_insuree.marital = ImisMaritalStatus.SINGLE.value
                    elif code == R4MaritalConfig.get_fhir_divorced_code():
                        imis_insuree.marital = ImisMaritalStatus.DIVORCED.value
                    elif code == R4MaritalConfig.get_fhir_widowed_code():
                        imis_insuree.marital = ImisMaritalStatus.WIDOWED.value
                    elif code == R4MaritalConfig.get_fhir_unknown_marital_status_code():
                        imis_insuree.marital = ImisMaritalStatus.NOT_SPECIFIED.value

    @classmethod
    def build_fhir_telecom(cls, fhir_patient, imis_insuree):
        fhir_patient.telecom = cls.build_fhir_telecom_for_person(phone=imis_insuree.phone, email=imis_insuree.email)

    @classmethod
    def build_imis_contacts(cls, imis_insuree, fhir_patient):
        imis_insuree.phone, imis_insuree.email = cls.build_imis_phone_num_and_email(fhir_patient.telecom)

    @classmethod
    def build_fhir_addresses(cls, fhir_patient, imis_insuree, reference_type):
        addresses = []
        if imis_insuree.family and imis_insuree.family.location:
            family_address = cls._build_insuree_family_address(imis_insuree.family, reference_type)
            addresses.append(family_address)

        if imis_insuree.current_village:
            # TODO: With current IG definition requiring location reference extension
            #  it's impossible to add current_address without current_village.
            insuree_address = cls._build_insuree_address(imis_insuree, reference_type)
            addresses.append(insuree_address)

        fhir_patient.address = addresses
        cls._validate_fhir_address_details(fhir_patient.address)

    @classmethod
    def build_imis_addresses(cls, imis_insuree, fhir_patient):
        cls._build_imis_current_patient_address(imis_insuree, fhir_patient)
        cls._build_or_validate_family_address(imis_insuree, fhir_patient)

    @classmethod
    def _build_imis_current_patient_address(cls, imis_insuree, fhir_patient):
        cls._validate_fhir_address_details(fhir_patient.address)
        if insuree_address := cls.__get_insuree_current_address_from_fhir_patinet(fhir_patient):
            imis_insuree.current_village = cls.__get_location_from_address(insuree_address)
            if insuree_address.text:
                imis_insuree.current_address = insuree_address.text

    @classmethod
    def _build_or_validate_family_address(cls, imis_insuree, fhir_patient):
        if not (fhir_family_address := cls.__get_family_address_from_fhir_patient(fhir_patient)):
            return

        if not (family := cls.__get_family_from_fhir_patient_extension(fhir_patient)):
            family_location = cls.__get_location_from_address(fhir_family_address)
            # Additional attribute for purpose of creating new family
            imis_insuree.family_location = family_location
            imis_insuree.family_address = fhir_family_address.text

        # Temporary disabled, this type of check shouldn't be part of converter.
        # if family and family.location and (family.location.uuid != family_location.uuid):
        #     raise FHIRException(
        #         f"Patient assigned to family {family} but provided family location {family_location} "
        #         f"does not match family location {family.location}")

    @classmethod
    def build_fhir_extentions(cls, fhir_patient, imis_insuree, reference_type):
        fhir_patient.extension = []

        def build_extension(fhir_patient, imis_insuree, value):
            extension = Extension.construct()
            if value == "head":
                extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-is-head"
                extension.valueBoolean = imis_insuree.head

            elif value == "education.education":
                extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-education-level"
                if hasattr(imis_insuree, "education") and imis_insuree.education is not None:
                    display = EducationLevelMapping.education_level[str(imis_insuree.education.id)]
                    system = "CodeSystem/patient-education-level"
                    extension.valueCodeableConcept = cls.build_codeable_concept(code=str(imis_insuree.education.id), system=system)
                    if len(extension.valueCodeableConcept.coding) == 1:
                        extension.valueCodeableConcept.coding[0].display = display

            elif value == "patient.card.issue":
                extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-card-issued"
                extension.valueBoolean = imis_insuree.card_issued

            elif value == "patient.group.reference":
                extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-group-reference"
                extension.valueReference = GroupConverter\
                    .build_fhir_resource_reference(imis_insuree.family, 'Group', reference_type=reference_type)

            elif value == "patient.identification":
                nested_extension = Extension.construct()
                extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-identification"
                if hasattr(imis_insuree, "type_of_id") and imis_insuree.type_of_id:
                    if hasattr(imis_insuree, "passport") and imis_insuree.passport:
                        # add number extension
                        nested_extension.url = "number"
                        nested_extension.valueString = imis_insuree.passport
                        extension.extension = [nested_extension]
                        # add identifier extension
                        nested_extension = Extension.construct()
                        nested_extension.url = "type"
                        system = "CodeSystem/patient-identification-type"
                        nested_extension.valueCodeableConcept = cls.build_codeable_concept(code=imis_insuree.type_of_id.code, system=system)
                        extension.extension.append(nested_extension)

            else:
                extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-profession"
                if hasattr(imis_insuree, "profession") and imis_insuree.profession is not None:
                    display = PatientProfessionMapping.patient_profession[str(imis_insuree.profession.id)]
                    system = "CodeSystem/patient-profession"
                    extension.valueCodeableConcept = cls.build_codeable_concept(code=str(imis_insuree.profession.id), system=system)
                    if len(extension.valueCodeableConcept.coding) == 1:
                        extension.valueCodeableConcept.coding[0].display = display

            if type(fhir_patient.extension) is not list:
                fhir_patient.extension = [extension]
            else:
                fhir_patient.extension.append(extension)


        if imis_insuree.head is not None:
            build_extension(fhir_patient, imis_insuree, "head")
        if imis_insuree.education is not None:
            build_extension(fhir_patient, imis_insuree, "education.education")
        if imis_insuree.profession is not None:
            build_extension(fhir_patient, imis_insuree, "profession.profession")
        if imis_insuree.card_issued is not None:
            build_extension(fhir_patient, imis_insuree, "patient.card.issue")
        if imis_insuree.family is not None:
            build_extension(fhir_patient, imis_insuree, "patient.group.reference")
        if imis_insuree.type_of_id is not None and imis_insuree.passport is not None:
            build_extension(fhir_patient, imis_insuree, "patient.identification")

    @classmethod
    def build_fhir_contact(cls, fhir_patient, imis_insuree):
        fhir_contact = PatientContact.construct()
        if imis_insuree.relationship is not None and imis_insuree.family is not None \
                and imis_insuree.family.head_insuree is not None:
            system = "CodeSystem/patient-contact-relationship"
            # map to the fhir value from imis one
            display = RelationshipMapping.relationship[str(imis_insuree.relationship.id)]
            fhir_contact.relationship = [cls.build_codeable_concept(code=imis_insuree.relationship.id, system=system)]
            fhir_contact.relationship[0].coding[0].display = display
            fhir_contact.name = cls.build_fhir_names_for_person(imis_insuree)

            if type(fhir_patient.contact) is not list:
                fhir_patient.contact = [fhir_contact]
            else:
                fhir_patient.contact.append(fhir_contact)

    @classmethod
    def build_fhir_photo(cls, fhir_patient, imis_insuree):
        if imis_insuree.photo and imis_insuree.photo.folder and imis_insuree.photo.filename:
            # HOST is taken from global variable used in the docker initialization
            # If URL root is not explicitly given in the settings 'localhost' is used
            # (if value is empty validation exception is raised).
            abs_url = GeneralConfiguration.get_host_domain().split('http://')[1] or 'localhost'
            domain = abs_url
            photo_uri = cls.__build_photo_uri(imis_insuree)
            photo = Attachment.construct()
            parsed = urllib.parse.urlunparse(('http', domain, photo_uri, None, None, None))
            photo.url = parsed
            photo.creation = imis_insuree.photo.date.isoformat()
            photo.contentType = imis_insuree.photo.filename[imis_insuree.photo.filename.rfind('.') + 1:]
            photo.title = imis_insuree.photo.filename
            if type(fhir_patient.photo) is not list:
                fhir_patient.photo = [photo]
            else:
                fhir_patient.photo.append(photo)

    @classmethod
    def build_imis_photo(cls, imis_insuree, fhir_patient, errors):
        if fhir_patient.photo and len(fhir_patient.photo) > 0:
            cls._validate_fhir_photo(fhir_patient)
            if fhir_patient.photo[0].data:
                photo = fhir_patient.photo[0].data
                date = fhir_patient.photo[0].creation
                obj, created = \
                    InsureePhoto.objects.get_or_create(
                        chf_id=imis_insuree.chf_id,
                        defaults={
                            "photo": photo,
                            "date": date,
                            "audit_user_id": -1,
                            "officer_id": 3
                        }
                    )
                imis_insuree.photo_id = obj.id

    @classmethod
    def build_fhir_general_practitioner(cls, fhir_patient, imis_insuree, reference_type):
        if imis_insuree.health_facility:
            hf = cls.build_fhir_resource_reference(
                imis_insuree.health_facility,
                'Organization',
                reference_type=reference_type
            )
            fhir_patient.generalPractitioner = [hf]

    @classmethod
    def build_imis_general_practitioner(cls, imis_insuree, fhir_patient):
        if not fhir_patient.generalPractitioner:
            return
        if len(fhir_patient.generalPractitioner) != 1:
            # TODO: Check with IG definition. Currently its 0..* but it's not possible to bind more than one HF
            raise FHIRException(_("Patient can provide at most one general practitioner."))

        hf_uuid = cls.get_resource_id_from_reference(fhir_patient.generalPractitioner[0])
        try:
            health_facility = HealthFacility.objects.get(uuid=hf_uuid)
            imis_insuree.health_facility = health_facility
        except HealthFacility.DoesNotExist:
            raise FHIRException(F"Invalid location reference, {hf_uuid} doesn't match any HealthFacility.")

    @classmethod
    def _family_reference_identifier_type(cls, reference_type):
        if reference_type == ReferenceConverterMixin.UUID_REFERENCE_TYPE:
            return cls.build_codeable_concept(R4IdentifierConfig.get_fhir_uuid_type_code())
        elif reference_type == ReferenceConverterMixin.DB_ID_REFERENCE_TYPE:
            return cls.build_codeable_concept(R4IdentifierConfig.get_fhir_id_type_code())
        elif reference_type == ReferenceConverterMixin.CODE_REFERENCE_TYPE:
            # Family don't have code assigned, uuid is used instead
            return cls.build_codeable_concept(R4IdentifierConfig.get_fhir_uuid_type_code())
        pass

    @classmethod
    def _family_reference_identifier_value(cls, family, reference_type):
        if reference_type == ReferenceConverterMixin.UUID_REFERENCE_TYPE:
            return family.uuid
        elif reference_type == ReferenceConverterMixin.DB_ID_REFERENCE_TYPE:
            return family.id
        elif reference_type == ReferenceConverterMixin.CODE_REFERENCE_TYPE:
            # Family don't have code assigned, uuid is used instead
            return family.uuid
        raise NotImplementedError(F"Reference type {reference_type} not implemented for family")

    @classmethod
    def __build_photo_uri(cls, imis_insuree):
        photo_folder = imis_insuree.photo.folder.replace("\\", "/")
        photo_full_path = F"{photo_folder}/{imis_insuree.photo.filename}"
        path = f'/photo/{photo_full_path}'
        return path

    @classmethod
    def _validate_fhir_identifier_is_exist(cls, fhir_patient):
        if not fhir_patient.identifier or len(fhir_patient.identifier) == 0:
            raise FHIRException(
                _('FHIR Patient entity without identifier')
            )

    # fhir validations
    @classmethod
    def _validate_fhir_extension_is_exist(cls, fhir_patient):
        if not fhir_patient.extension or len(fhir_patient.extension) == 0:
            raise FHIRException(
                _('At least one extension with is_head is required')
            )

    @classmethod
    def _validate_imis_is_head(cls, imis_insuree):
        if imis_insuree.head is None:
            raise FHIRException(
                _('Missing is-head in IMIS object')
            )

    @classmethod
    def _validate_fhir_address(cls, fhir_family):
        if not fhir_family.address or len(fhir_family.address) == 0:
            raise FHIRException(
                _('Address must be supported')
            )

    @classmethod
    def _validate_fhir_address_details(cls, addresses):
        addr_errors = {}
        for idx, address in enumerate(addresses):
            errors = []
            # Get last part of each extension url
            if not address.extension:
                raise FHIRException("Missing extensions for Address")

            ext_types = [ext.url.rsplit('/')[-1] for ext in address.extension]
            if 'address-location-reference' not in ext_types:
                errors.append("FHIR Patient address without address-location-reference extension.")
            if 'address-municipality' not in ext_types:
                errors.append("FHIR Patient address without address-municipality reference.")
            if len(ext_types) != 2:
                errors.append("Patient's address should provide exactly 2 extensions")
            if not address.city:
                errors.append("Address 'city' field required")
            if not address.district:
                errors.append("Address 'district' field required")
            if not address.state:
                errors.append("Address 'state' field required")

            if errors:
                addr_errors[idx] = errors

        if addr_errors:
            raise FHIRException(json.dumps(addr_errors))

    @classmethod
    def _validate_fhir_photo(cls, fhir_patient):
        if not fhir_patient.photo or len(fhir_patient.photo) == 0:
            raise FHIRException(
                _('FHIR Patient without photo data.')
            )
        else:
            photo = fhir_patient.photo[0]
            if not photo.title or not photo.creation or not photo.contentType:
                raise FHIRException(
                    _('FHIR Patient misses one of required fields:  contentType, title, creation')
                )

    # imis validations
    @classmethod
    def _validate_imis_identifier_code(cls, imis_insuree):
        if not imis_insuree.chf_id:
            raise FHIRException(
                _('Insuree %(insuree_uuid)s without code') % {'insuree_uuid': imis_insuree.uuid}
            )

    @classmethod
    def _validate_fhir_family_home_slice(cls, fhir_patient):
        is_home_use = False
        for address in fhir_patient.address:
            if address.use == "home":
                is_home_use = True
        if is_home_use == False:
            raise FHIRException(
                _('Patient without family address')
            )

    @classmethod
    def _validate_fhir_patient_human_name(cls, fhir_patient):
        if not fhir_patient.name:
            raise FHIRException(
                _('Missing fhir patient attribute: name')
            )
        for name in fhir_patient.name:
            if not name.family or not name.given:
                raise FHIRException(
                    _('Missing obligatory fields for fhir patient name: family or given')
                )

    @classmethod
    def _validate_imis_insuree_human_name(cls, imis_insuree):
        if not imis_insuree.last_name or not imis_insuree.other_names:
            raise FHIRException(
                _('Missing patient family name or given name')
            )

    @classmethod
    def __add_insuree_current_address(cls, imis_insuree, address):
        imis_insuree.current_address = address.text

    @classmethod
    def __add_insuree_location(cls, imis_insuree, address):
        # Physical address can also provide information regarding current address
        if address.text:
            imis_insuree.current_address = address.text

        for ext in address.extension:
            if "StructureDefinition/address-location-reference" in ext.url:
                location_uuid = LocationConverter.get_resource_id_from_reference(ext.valueReference)
                try:
                    location = Location.objects.get(uuid=location_uuid)
                    imis_insuree.current_village = location
                except Location.DoesNotExist as e:
                    raise FHIRException(f"Invalid location reference, {location_uuid} doesn't match any location.")

    @classmethod
    def __add_insuree_geolocation(cls, imis_insuree, address):
        imis_insuree.geolocation = address.text

    @classmethod
    def _build_insuree_family_address(cls, imis_insuree_family: Family, reference_type):
        return cls.__build_address_of_use(
            address_location=imis_insuree_family.location,
            use='home',
            location_text=imis_insuree_family.address,
            reference_type=reference_type
        )

    @classmethod
    def _build_insuree_address(cls, imis_insuree, reference_type):
        return cls.__build_address_of_use(
            address_location=imis_insuree.current_village,
            use='temp',
            location_text=imis_insuree.current_address,
            reference_type=reference_type
        )

    @classmethod
    def __build_address_of_use(cls, address_location: Location, use: str, location_text: str, reference_type):
        base_address = cls.__build_base_physical_address(address_location, reference_type)
        base_address.use = use
        if location_text:
            base_address.text = location_text
        return base_address

    @classmethod
    def __build_base_physical_address(cls, imis_village_location, reference_type) -> Address:
        return Address(**{
            "type": "physical",
            "state": cls.__state_name_from_physical_location(imis_village_location),
            "district": cls.__district_name_from_physical_location(imis_village_location),
            "city": cls.__village_name_from_physical_location(imis_village_location),
            "extension": [
                cls.__build_municipality_extension(imis_village_location),
                cls.__build_location_reference_extension(imis_village_location, reference_type)
            ]
        })

    @classmethod
    def __build_municipality_extension(cls, insuree_family_location):
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-municipality"
        extension.valueString = cls.__municipality_from_family_location(insuree_family_location)
        return extension

    @classmethod
    def __state_name_from_physical_location(cls, insuree_family_location):
        return insuree_family_location.parent.parent.parent.name

    @classmethod
    def __district_name_from_physical_location(cls, insuree_family_location):
        return insuree_family_location.parent.parent.name

    @classmethod
    def __municipality_from_family_location(cls, insuree_family_location):
        return insuree_family_location.parent.name

    @classmethod
    def __village_name_from_physical_location(cls, insuree_family_location):
        return insuree_family_location.name

    @classmethod
    def __build_location_reference_extension(cls, insuree_family_location, reference_type):
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-location-reference"
        extension.valueReference = LocationConverter \
            .build_fhir_resource_reference(insuree_family_location, 'Location', reference_type=reference_type)
        return extension

    @classmethod
    def __build_imis_family_address(cls, fhir_patient):
        family_address = cls.__get_family_address_from_fhir_patient(fhir_patient)
        return cls.__get_location_from_address(family_address) if family_address else None

    @classmethod
    def __get_location_from_address(cls, fhir_patient_address):
        try:
            location_reference = next((
                ext for ext in fhir_patient_address.extension if 'address-location-reference' in ext.url
            ))
            location_uuid = LocationConverter.get_resource_id_from_reference(location_reference.valueReference)
            return Location.objects.get(uuid=location_uuid)
        except Location.DoesNotExist as e:
            raise FHIRException(f"Invalid location reference, {location_uuid} doesn't match any location.")

    @classmethod
    def __get_family_address_from_fhir_patient(cls, fhir_patient) -> Address:
        # Family address is of use home and type physical
        return cls.__get_address_by_properties(fhir_patient, 'home', 'physical')

    @classmethod
    def __get_insuree_current_address_from_fhir_patinet(cls, fhir_patient) -> Address:
        # Family address is of use temp and type physical
        return cls.__get_address_by_properties(fhir_patient, 'temp', 'physical')

    @classmethod
    def __get_address_by_properties(cls, fhir_patient, use, type_):
        matching = [
            addr for addr in fhir_patient.address if addr.use == use and addr.type == type_
        ]
        if len(matching) > 1:
            raise FHIRException(F"More than one address with use {use} and type {type_} provided")

        return matching[0] if matching else None

    @classmethod
    def __get_family_from_fhir_patient_extension(cls, fhir_patient):
        if not (family_uuid := cls.__get_family_uuid_from_fhir_patient(fhir_patient)):
            return None

        try:
            return Family.objects.select_related('location').get(uuid=family_uuid)
        except Family.DoesNotExist as e:
            raise FHIRException(F"Invalid location reference, {e} doesn't match any location.")

    @classmethod
    def __get_family_uuid_from_fhir_patient(cls, fhir_patient):
        return next((
            GroupConverter.get_resource_id_from_reference(ext.valueReference)
            for ext in fhir_patient.extension if "/StructureDefinition/patient-group-reference" in ext.url
        ), None)



from typing import Dict, Type

from api_fhir_r4.containedResources.containedResources import AbstractContainedResourceCollection, \
    ContainedResourceDefinition
from api_fhir_r4.serializers import BaseFHIRSerializer, PatientSerializer, GroupSerializer, \
    HealthFacilityOrganisationSerializer, ClaimAdminPractitionerSerializer, MedicationSerializer, \
    ActivityDefinitionSerializer, ClaimAdminPractitionerRoleSerializer


class ClaimContainedResources(AbstractContainedResourceCollection):
    @classmethod
    def _definitions_for_serializers(cls) -> Dict[Type[BaseFHIRSerializer], ContainedResourceDefinition]:
        return {
            PatientSerializer: ContainedResourceDefinition('insuree', 'Patient'),
            GroupSerializer: ContainedResourceDefinition(
                'insuree', 'Group',
                lambda model, field: model.__getattribute__(field).family
            ),
            HealthFacilityOrganisationSerializer: ContainedResourceDefinition('health_facility', 'Organization'),
            ClaimAdminPractitionerSerializer: ContainedResourceDefinition('admin', 'Practitioner'),
            ClaimAdminPractitionerRoleSerializer: ContainedResourceDefinition('admin', 'PractitionerRole'),
            MedicationSerializer: ContainedResourceDefinition(
                'items', 'Medication',
                lambda model, field: [
                    item.item for item in model.__getattribute__(field).filter(validity_to=None).all()
                ]
            ),
            ActivityDefinitionSerializer: ContainedResourceDefinition(
                'services', 'ActivityDefinition',
                lambda model, field: [
                    service.service for service in model.__getattribute__(field).filter(validity_to=None).all()
                ]
            ),
        }



from rest_framework.test import APITestCase
from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiCreateTestMixin, FhirApiReadTestMixin, \
    FhirApiDeleteTestMixin, FhirApiUpdateTestMixin
from api_fhir_r4.tests.mixin.SubscriptionTestMixin import SubscriptionTestMixin
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin


class SubscriptionAPITests(GenericFhirAPITestMixin, APITestCase, LogInMixin, FhirApiCreateTestMixin,
                           FhirApiReadTestMixin, FhirApiDeleteTestMixin, FhirApiUpdateTestMixin,
                           SubscriptionTestMixin):
    base_url = GeneralConfiguration.get_base_url() + 'Subscription/'
    _test_json_path = "/test/test_subscription.json"
    _updated_endpoint = 'https://modifiedendpoint.com'

    def update_resource(self, data):
        data['channel']['endpoint'] = self._updated_endpoint

    def verify_updated_obj(self, obj):
        self.assertEqual(obj.channel.endpoint, self._updated_endpoint)

    def setUp(self):
        super(SubscriptionAPITests, self).setUp()
        user = self.get_or_create_user_api()
        subscription = self.create_test_imis_instance()
        subscription.id = None
        subscription.save(username=user.username)

    def get_id_for_created_resource(self, response):
        return response.data['id']



from api_fhir_r4.configurations import GeneralConfiguration


class ItemTypeMapping(object):
    item_type = {
        "D": "Drug",
        "M": "Medical_Consumable",
    }


class ItemContextlevel:
    SYSTEM = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/medication-level"

    item_context_level_coding = {
        "system": SYSTEM,
        "code": "M",
        "display": "Medication",
    }


class ItemVenueTypeMapping(object):
    item_venue_type = {
        "AMB": "ambulatory",
        "IMP": "IMP",
        "B": "both",
    }

    venue_fhir_imis = {
        "AMB": "O",
        "IMP": "I",
    }



import datetime

from django.db.models import Prefetch
from rest_framework import mixins
from rest_framework.response import Response
from rest_framework.serializers import ValidationError
from rest_framework.viewsets import GenericViewSet

from api_fhir_r4.mixins import MultiIdentifierRetrieverMixin
from api_fhir_r4.model_retrievers import UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever
from api_fhir_r4.permissions import FHIRApiClaimPermissions
from api_fhir_r4.serializers import ClaimSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from claim.models import Claim, ClaimItem, ClaimService
from insuree.models import Insuree, InsureePolicy


class ClaimViewSet(BaseFHIRView, MultiIdentifierRetrieverMixin, mixins.ListModelMixin,
                   mixins.CreateModelMixin, GenericViewSet):
    retrievers = [UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever]
    serializer_class = ClaimSerializer
    permission_classes = (FHIRApiClaimPermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        refDate = request.GET.get('refDate')
        identifier = request.GET.get("identifier")
        patient = request.GET.get("patient")
        contained = bool(request.GET.get("contained"))

        if identifier is not None:
            return self.retrieve(request, *args, **{**kwargs, 'identifier': identifier})
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by('validity_from')
            if refDate is not None:
                try:
                    date_from = datetime.datetime.strptime(refDate, "%Y-%m-%d").date()
                    queryset = queryset.filter(validity_from__gte=date_from)
                except ValueError as v:
                    raise ValidationError({'refDate': 'Invalid date format, should be in "%Y-%m-%d" format'})

            if patient is not None:
                for_patient = Insuree.objects.get(uuid=patient)
                queryset = queryset.filter(insuree=for_patient)

        serializer = ClaimSerializer(self.paginate_queryset(queryset), many=True, context={'contained': contained})
        return self.get_paginated_response(serializer.data)

    def retrieve(self, request, *args, **kwargs):
        contained = bool(request.GET.get("contained"))
        ref_type, instance = self._get_object_with_first_valid_retriever(kwargs['identifier'])
        serializer = self.get_serializer(instance, context={'contained': contained}, reference_type=ref_type)
        return Response(serializer.data)

    def get_queryset(self):
        queryset = Claim.get_queryset(None, self.request.user).order_by('validity_from') \
            .select_related('insuree') \
            .select_related('health_facility') \
            .select_related('icd') \
            .select_related('icd_1') \
            .select_related('icd_2') \
            .select_related('icd_3') \
            .select_related('icd_4') \
            .prefetch_related(Prefetch('items', queryset=ClaimItem.objects.filter(validity_to__isnull=True))) \
            .prefetch_related(Prefetch('services', queryset=ClaimService.objects.filter(validity_to__isnull=True))) \
            .prefetch_related(Prefetch('insuree__insuree_policies',
                                       queryset=InsureePolicy.objects.filter(validity_to__isnull=True).select_related(
                                           "policy")))
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(queryset)



from api_fhir_r4.converters import ClaimAdminPractitionerRoleConverter
from fhir.resources.practitionerrole import PractitionerRole
from api_fhir_r4.tests import ClaimAdminPractitionerRoleTestMixin
from api_fhir_r4.tests.mixin import ConvertJsonToFhirTestMixin, ConvertToFhirTestMixin, ConvertToImisTestMixin


class ClaimAdminPractitionerRoleConverterTestCase(ClaimAdminPractitionerRoleTestMixin,
                                                  ConvertToImisTestMixin,
                                                  ConvertToFhirTestMixin,
                                                  ConvertJsonToFhirTestMixin):
    converter = ClaimAdminPractitionerRoleConverter
    fhir_resource = PractitionerRole
    json_repr = 'test/test_claimAdminPractitionerRole.json'



from core.models import Officer
from django.utils.translation import gettext as _

from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin
from fhir.resources.practitioner import Practitioner, PractitionerQualification
from api_fhir_r4.utils import TimeUtils, DbManagerUtils


class EnrolmentOfficerPractitionerConverter(BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_officer, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_practitioner = Practitioner.construct()
        cls.build_fhir_pk(fhir_practitioner, imis_officer, reference_type)
        cls.build_fhir_identifiers(fhir_practitioner, imis_officer)
        cls.build_human_names(fhir_practitioner, imis_officer)
        cls.build_fhir_birth_date(fhir_practitioner, imis_officer)
        cls.build_fhir_telecom(fhir_practitioner, imis_officer)
        cls.build_fhir_qualification(fhir_practitioner)
        return fhir_practitioner

    @classmethod
    def to_imis_obj(cls, fhir_practitioner, audit_user_id):
        errors = []
        fhir_practitioner = Practitioner(**fhir_practitioner)
        imis_officer = EnrolmentOfficerPractitionerConverter.create_default_claim_admin(audit_user_id)
        cls.build_imis_identifiers(imis_officer, fhir_practitioner, errors)
        cls.build_imis_names(imis_officer, fhir_practitioner)
        cls.build_imis_birth_date(imis_officer, fhir_practitioner)
        cls.build_imis_contacts(imis_officer, fhir_practitioner)
        cls.check_errors(errors)
        return imis_officer

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def get_reference_obj_uuid(cls, imis_officer: Officer):
        return imis_officer.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_officer: Officer):
        return imis_officer.id

    @classmethod
    def get_reference_obj_code(cls, imis_officer: Officer):
        return imis_officer.code

    @classmethod
    def get_fhir_resource_type(cls):
        return Practitioner

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_officer_uuid = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Officer, uuid=imis_officer_uuid)

    @classmethod
    def create_default_claim_admin(cls, audit_user_id):
        imis_officer = Officer()
        imis_officer.validity_from = TimeUtils.now()
        imis_officer.audit_user_id = audit_user_id
        return imis_officer

    @classmethod
    def build_fhir_identifiers(cls, fhir_practitioner, imis_officer):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_officer)
        fhir_practitioner.identifier = identifiers

    @classmethod
    def build_fhir_code_identifier(cls, identifiers, imis_officer):
        if imis_officer.code:
            identifier = cls.build_fhir_identifier(imis_officer.code,
                                                   R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                   R4IdentifierConfig.get_fhir_generic_type_code())
            identifiers.append(identifier)

    @classmethod
    def build_imis_identifiers(cls, imis_officer, fhir_practitioner, errors):
        value = cls.get_fhir_identifier_by_code(fhir_practitioner.identifier,
                                                R4IdentifierConfig.get_fhir_generic_type_code())
        if value:
            imis_officer.code = value
        cls.valid_condition(imis_officer.code is None, _('Missing the claim admin code'), errors)

    @classmethod
    def build_human_names(cls, fhir_practitioner, imis_officer):
        name = cls.build_fhir_names_for_person(imis_officer)
        fhir_practitioner.name = [name]

    @classmethod
    def build_imis_names(cls, imis_officer, fhir_practitioner):
        names = fhir_practitioner.name
        imis_officer.last_name, imis_officer.other_names = cls.build_imis_last_and_other_name(names)

    @classmethod
    def build_fhir_birth_date(cls, fhir_practitioner, imis_officer):
        if imis_officer.dob is not None:
            from core import datetime
            # check if datetime object
            if isinstance(imis_officer.dob, datetime.datetime):
                fhir_practitioner.birthDate = imis_officer.dob.date().isoformat()
            else:
                fhir_practitioner.birthDate = imis_officer.dob.isoformat()

    @classmethod
    def build_imis_birth_date(cls, imis_officer, fhir_practitioner):
        birth_date = fhir_practitioner.birthDate
        if birth_date:
            imis_officer.dob = TimeUtils.str_to_date(birth_date)

    @classmethod
    def build_fhir_telecom(cls, fhir_practitioner, imis_officer):
        fhir_practitioner.telecom = cls.build_fhir_telecom_for_person(phone=imis_officer.phone,
                                                                      email=imis_officer.email)

    @classmethod
    def build_imis_contacts(cls, imis_officer, fhir_practitioner):
        imis_officer.phone, imis_officer.email = cls.build_imis_phone_num_and_email(fhir_practitioner.telecom)

    @classmethod
    def build_fhir_qualification(cls, fhir_practitioner):
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/practitioner-qualification-type"
        qualification = PractitionerQualification.construct()
        qualification.code = cls.build_codeable_concept(
            system=system,
            code="EO",
            display=_("Enrolment Officer")
        )
        fhir_practitioner.qualification = [qualification]



from api_fhir_r4.converters import EnrolmentOfficerPractitionerRoleConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class EnrolmentOfficerPractitionerRoleSerializer(BaseFHIRSerializer):

    fhirConverter = EnrolmentOfficerPractitionerRoleConverter()



from urllib.parse import urljoin

from django.utils.translation import gettext as _
from fhir.resources.address import Address
from fhir.resources.extension import Extension
from fhir.resources.humanname import HumanName

from api_fhir_r4.mapping.organizationMapping import PolicyHolderOrganisationLegalFormMapping, \
    PolicyHolderOrganisationActivityMapping
from api_fhir_r4.models.imisModelEnums import ImisLocationType, ContactPointSystem, AddressType
from location.models import Location
from policyholder.models import PolicyHolder
from api_fhir_r4.configurations import R4IdentifierConfig, R4OrganisationConfig, GeneralConfiguration
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from fhir.resources.organization import Organization
from api_fhir_r4.utils import DbManagerUtils


class PolicyHolderOrganisationConverter(BaseFHIRConverter, ReferenceConverterMixin):
    @classmethod
    def to_fhir_obj(cls, imis_organisation, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_organisation = Organization()
        cls.build_fhir_pk(fhir_organisation, imis_organisation, reference_type)
        cls.build_fhir_extensions(fhir_organisation, imis_organisation)
        cls.build_fhir_identifiers(fhir_organisation, imis_organisation, reference_type)
        cls.build_fhir_type(fhir_organisation)
        cls.build_fhir_name(fhir_organisation, imis_organisation)
        cls.build_fhir_telecom(fhir_organisation, imis_organisation)
        cls.build_fhir_ph_address(fhir_organisation, imis_organisation, reference_type)
        cls.build_fhir_contact(fhir_organisation, imis_organisation)
        return fhir_organisation

    @classmethod
    def to_imis_obj(cls, fhir_organisation, audit_user_id):
        raise NotImplementedError(
            _('PH Organization to_imis_obj() not implemented.')
        )

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_ph_uuid = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(PolicyHolder, uuid=imis_ph_uuid)

    @classmethod
    def get_reference_obj_id(cls, obj):
        return obj.uuid

    @classmethod
    def get_reference_obj_uuid(cls, obj):
        return obj.uuid

    @classmethod
    def get_reference_obj_code(cls, obj):
        return obj.code

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def get_fhir_resource_type(cls):
        return Organization

    @classmethod
    def build_fhir_identifiers(cls, fhir_organisation, imis_organisation, reference_type):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_organisation, reference_type)
        fhir_organisation.identifier = identifiers

    @classmethod
    def build_fhir_extensions(cls, fhir_organisation, imis_organisation):
        cls.build_fhir_legal_form_extension(fhir_organisation, imis_organisation)
        cls.build_fhir_activity_extension(fhir_organisation, imis_organisation)

    @classmethod
    def build_fhir_legal_form_extension(cls, fhir_organisation, imis_organisation):
        codeable_concept = cls.build_codeable_concept_from_coding(cls.build_fhir_mapped_coding(
            PolicyHolderOrganisationLegalFormMapping.fhir_ph_code_system(imis_organisation.legal_form)
        ))
        base = GeneralConfiguration.get_system_base_url()
        url = urljoin(base, R4OrganisationConfig.get_fhir_ph_organisation_legal_form_extension_system())
        extension = cls.build_fhir_codeable_concept_extension(codeable_concept, url)
        if isinstance(fhir_organisation.extension, list):
            fhir_organisation.extension.append(extension)
        else:
            fhir_organisation.extension = [extension]

    @classmethod
    def build_fhir_activity_extension(cls, fhir_organisation, imis_organisation):
        codeable_concept = cls.build_codeable_concept_from_coding(cls.build_fhir_mapped_coding(
            PolicyHolderOrganisationActivityMapping.fhir_ph_code_system(imis_organisation.activity_code)
        ))
        base = GeneralConfiguration.get_system_base_url()
        url = urljoin(base, R4OrganisationConfig.get_fhir_ph_organisation_activity_extension_system())
        extension = cls.build_fhir_codeable_concept_extension(codeable_concept, url)
        if isinstance(fhir_organisation.extension, list):
            fhir_organisation.extension.append(extension)
        else:
            fhir_organisation.extension = [extension]

    @classmethod
    def build_fhir_type(cls, fhir_organisation):
        fhir_organisation.type = [cls.build_codeable_concept(
            R4OrganisationConfig.get_fhir_ph_organisation_type(),
            system=R4OrganisationConfig.get_fhir_ph_organisation_type_system()
        )]

    @classmethod
    def build_fhir_name(cls, fhir_organisation, imis_organisation):
        fhir_organisation.name = imis_organisation.trade_name

    @classmethod
    def build_fhir_telecom(cls, fhir_organisation, imis_organisation):
        fhir_organisation.telecom = []
        if imis_organisation.email:
            fhir_organisation.telecom.append(cls.build_fhir_contact_point(
                system=ContactPointSystem.EMAIL,
                value=imis_organisation.email))
        if imis_organisation.fax:
            fhir_organisation.telecom.append(cls.build_fhir_contact_point(
                system=ContactPointSystem.FAX,
                value=imis_organisation.fax))
        if imis_organisation.phone:
            fhir_organisation.telecom.append(cls.build_fhir_contact_point(
                system=ContactPointSystem.PHONE,
                value=imis_organisation.phone))

    @classmethod
    def build_fhir_ph_address(cls, fhir_organisation, imis_organisation, reference_type):
        address = Address.construct()
        address.type = AddressType.PHYSICAL.value
        if imis_organisation.address and "address" in imis_organisation.address:
            address.line = [imis_organisation.address["address"]]
        fhir_organisation.address = [address]
        if imis_organisation.locations:
            cls.build_fhir_address_field(fhir_organisation, imis_organisation.locations)
            cls.build_fhir_location_extension(fhir_organisation, imis_organisation, reference_type)

    @classmethod
    def build_fhir_address_field(cls, fhir_organisation, location: Location):
        current_location = location
        while current_location:
            if current_location.type == ImisLocationType.REGION.value:
                fhir_organisation.address[0].state = current_location.name
            elif current_location.type == ImisLocationType.DISTRICT.value:
                fhir_organisation.address[0].district = current_location.name
            elif current_location.type == ImisLocationType.WARD.value:
                cls.build_fhir_municipality_extension(fhir_organisation, current_location)
            elif current_location.type == ImisLocationType.VILLAGE.value:
                fhir_organisation.address[0].city = current_location.name
            current_location = current_location.parent

    @classmethod
    def build_fhir_municipality_extension(cls, fhir_organisation, municipality: Location):
        extension = Extension.construct()
        base = GeneralConfiguration.get_system_base_url()
        extension.url = urljoin(base, R4OrganisationConfig.get_fhir_address_municipality_extension_system())
        extension.valueString = municipality.name
        if isinstance(fhir_organisation.address[0].extension, list):
            fhir_organisation.address[0].extension.append(extension)
        else:
            fhir_organisation.address[0].extension = [extension]

    @classmethod
    def build_fhir_location_extension(cls, fhir_organisation, imis_organisation, reference_type):
        base = GeneralConfiguration.get_system_base_url()
        url = urljoin(base, R4OrganisationConfig.get_fhir_location_reference_extension_system())
        extension = cls.build_fhir_reference_extension(cls.build_fhir_resource_reference
                                                       (imis_organisation.locations,
                                                        type='Location',
                                                        display=imis_organisation.locations.name,
                                                        reference_type=reference_type),
                                                       url)
        if isinstance(fhir_organisation.address[0].extension, list):
            fhir_organisation.address[0].extension.append(extension)
        else:
            fhir_organisation.address[0].extension = [extension]

    @classmethod
    def build_fhir_contact(cls, fhir_organisation, imis_organisation):
        fhir_organisation.contact = []
        if imis_organisation.contact_name:
            name = HumanName.construct()
            name.text = "%s %s" % (imis_organisation.contact_name['name'], imis_organisation.contact_name['surname'])
            fhir_organisation.contact.append({'name': name})



from django.utils.translation import gettext as _
from medical.models import Item
from api_fhir_r4.converters import R4IdentifierConfig, BaseFHIRConverter, ReferenceConverterMixin
from api_fhir_r4.models import UsageContextV2 as UsageContext
from api_fhir_r4.mapping.medicationMapping import ItemTypeMapping, ItemVenueTypeMapping, ItemContextlevel
from api_fhir_r4.mapping.patientMapping import PatientCategoryMapping
from fhir.resources.medication import Medication as FHIRMedication
from fhir.resources.extension import Extension
from fhir.resources.money import Money
from fhir.resources.codeableconcept import CodeableConcept
from fhir.resources.coding import Coding
from fhir.resources.quantity import Quantity
from fhir.resources.ratio import Ratio
from fhir.resources.timing import Timing, TimingRepeat
from django.utils.translation import gettext
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.utils import DbManagerUtils
from api_fhir_r4.configurations import GeneralConfiguration
import core


class MedicationConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_medication, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_medication = FHIRMedication.construct()
        cls.build_fhir_pk(fhir_medication, imis_medication, reference_type)
        cls.build_fhir_identifiers(fhir_medication, imis_medication)
        cls.build_fhir_package_form(fhir_medication, imis_medication)
        cls.build_fhir_package_amount(fhir_medication, imis_medication)
        cls.build_fhir_medication_extension(fhir_medication, imis_medication)
        cls.build_fhir_code(fhir_medication, imis_medication)
        cls.build_fhir_status(fhir_medication, imis_medication)
        cls.build_fhir_level(fhir_medication, imis_medication)
        return fhir_medication

    @classmethod
    def to_imis_obj(cls, fhir_medication, audit_user_id):
        errors = []
        fhir_medication = FHIRMedication(**fhir_medication)
        imis_medication = Item()
        imis_medication.audit_user_id = audit_user_id
        cls.build_imis_identifier(imis_medication, fhir_medication, errors)
        cls.build_imis_item_code(imis_medication, fhir_medication, errors)
        cls.build_imis_item_name(imis_medication, fhir_medication, errors)
        cls.build_imis_item_package(imis_medication, fhir_medication, errors)
        cls.build_imis_item_extension(imis_medication, fhir_medication, errors)
        cls.check_errors(errors)
        return imis_medication

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def get_reference_obj_uuid(cls, imis_medication: Item):
        return imis_medication.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_medication: Item):
        return imis_medication.id

    @classmethod
    def get_reference_obj_code(cls, imis_medication: Item):
        return imis_medication.code

    @classmethod
    def get_fhir_resource_type(cls):
        return FHIRMedication

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_medication_code = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Item, uuid=imis_medication_code)

    @classmethod
    def build_fhir_identifiers(cls, fhir_medication, imis_medication):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_medication)
        fhir_medication.identifier = identifiers

    @classmethod
    def build_imis_identifier(cls, imis_medication, fhir_medication, errors):
        value = cls.get_fhir_identifier_by_code(fhir_medication.identifier,
                                                R4IdentifierConfig.get_fhir_uuid_type_code())
        if value:
            imis_medication.code = value
        cls.valid_condition(imis_medication.code is None, gettext('Missing the item code'), errors)

    @classmethod
    def build_fhir_package_form(cls, fhir_medication, imis_medication):
        # TODO - Split medical item ItemPackage into ItemForm and ItemAmount => openIMIS side
        if imis_medication.package:
            codeable = CodeableConcept.construct()
            codeable.text = imis_medication.package.lstrip()
            fhir_medication.form = codeable

    @classmethod
    def split_package_form(cls, form):
        form = form.lstrip()
        if " " not in form:
            return form
        if " " in form:
            form = form.split(' ', 1)
            form = form[1]
            return form

    @classmethod
    def build_fhir_package_amount(cls, fhir_medication, imis_medication):
        # TODO - Split medical item ItemPackage into ItemForm and ItemAmount => openIMIS side
        if imis_medication.package:
            amount = cls.split_package_amount(imis_medication.package)
            ratio = Ratio.construct()
            numerator = Quantity.construct()
            numerator.value = amount
            ratio.numerator = numerator
            fhir_medication.amount = ratio

    @classmethod
    def split_package_amount(cls, amount):
        amount = amount.lstrip()
        if " " not in amount:
            return None
        if " " in amount:
            amount = amount.split(' ', 1)
            amount = amount[0]
            return int(amount)

    @classmethod
    def build_fhir_medication_extension(cls, fhir_medication, imis_medication):
        cls.build_fhir_unit_price(fhir_medication, imis_medication)
        cls.build_fhir_medication_type(fhir_medication, imis_medication)
        cls.build_fhir_medication_frequency(fhir_medication, imis_medication)
        cls.build_fhir_use_context(fhir_medication, imis_medication)

    @classmethod
    def build_fhir_unit_price(cls, fhir_medication, imis_medication):
        unit_price = cls.build_fhir_unit_price_extension(imis_medication.price)
        if type(fhir_medication.extension) is not list:
            fhir_medication.extension = [unit_price]
        else:
            fhir_medication.extension.append(unit_price)

    @classmethod
    def build_fhir_unit_price_extension(cls, value):
        extension = Extension.construct()
        money = Money.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/unit-price"
        extension.valueMoney = money
        extension.valueMoney.value = value
        if hasattr(core, 'currency'):
            extension.valueMoney.currency = core.currency
        return extension

    @classmethod
    def build_fhir_medication_type(cls, fhir_medication, imis_medication):
        medication_type = cls.build_fhir_medication_type_extension(imis_medication.type)
        if type(fhir_medication.extension) is not list:
            fhir_medication.extension = [medication_type]
        else:
            fhir_medication.extension.append(medication_type)

    @classmethod
    def build_fhir_medication_type_extension(cls, value):
        extension = Extension.construct()
        display = ItemTypeMapping.item_type[value]
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/medication-item-type"
        coding = cls.build_codeable_concept(code=value, system=system, display=_(display))
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/medication-type"
        extension.valueCodeableConcept = coding
        return extension

    @classmethod
    def build_fhir_medication_frequency(cls, fhir_medication, imis_medication):
        if imis_medication.frequency:
            medication_frequency = cls.build_fhir_medication_frequency_extension(imis_medication.frequency)
            if type(fhir_medication.extension) is not list:
                fhir_medication.extension = [medication_frequency]
            else:
                fhir_medication.extension.append(medication_frequency)

    @classmethod
    def build_fhir_medication_frequency_extension(cls, value):
        # TODO: Is this ok? Value is assigned to period instead of frequency
        extension = Extension.construct()
        timing = Timing.construct()
        timing_repeat = TimingRepeat.construct()
        timing_repeat.frequency = 1
        timing_repeat.period = str(value)
        timing_repeat.periodUnit = 'd'
        timing.repeat = timing_repeat
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/medication-frequency"
        extension.valueTiming = timing
        return extension

    @classmethod
    def build_fhir_code(cls, fhir_medication, imis_medication):
        codeable = CodeableConcept.construct()
        codeable.text = imis_medication.name
        fhir_medication.code = codeable

    @classmethod
    def build_fhir_status(cls, fhir_medication, imis_medication):
        fhir_medication.status = "active"

    @classmethod
    def build_imis_item_code(cls, imis_medication, fhir_medication, errors):
        item_code = cls.get_fhir_identifier_by_code(fhir_medication.identifier,
                                                    R4IdentifierConfig.get_fhir_generic_type_code())
        if not cls.valid_condition(item_code is None,
                                   gettext('Missing medication `item_code` attribute'), errors):
            imis_medication.code = item_code

    @classmethod
    def build_imis_item_name(cls, imis_medication, fhir_medication, errors):
        if not cls.valid_condition(fhir_medication.code is None,
                                   gettext('Missing medication `item_name` attribute'), errors):
            item_name = fhir_medication.code.text
            imis_medication.name = item_name

    @classmethod
    def build_imis_item_package(cls, imis_medication, fhir_medication, errors):
        if not cls.valid_condition(fhir_medication.form is None,
                                   gettext('Missing medication `form` and `amount` attribute'), errors):
            form = fhir_medication.form.text
            imis_medication.package = form

    @classmethod
    def build_imis_item_extension(cls, imis_medication, fhir_medication, errors):
        extensions = fhir_medication.extension
        for extension in extensions:
            if "unit-price" in extension.url:
                cls.build_imis_unit_price(imis_medication, extension)
            if "medication-type" in extension.url:
                cls.build_imis_medication_type(imis_medication, extension)
            if "medication-frequency" in extension.url:
                cls.build_imis_medication_frequency(imis_medication, extension)
            if "medication-usage-context" in extension.url:
                cls.build_imis_item_usage_context(imis_medication, extension)

    @classmethod
    def build_fhir_use_context(cls, fhir_medication, imis_medication):
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/medication-usage-context"
        gender = cls.build_fhir_gender(imis_medication)
        # check only the first to be sure if we have list, the
        # next ones for sure will be a part of list of extensions
        if type(extension.extension) is not list:
            extension.extension = [gender]
        else:
            extension.extension.append(gender)

        age = cls.build_fhir_age(imis_medication)
        extension.extension.append(age)

        care_type = cls.build_fhir_care_type(imis_medication)
        if care_type:
            extension.extension.append(care_type)
        fhir_medication.extension.append(extension)

    @classmethod
    def build_fhir_gender(cls, imis_medication):
        male_flag = PatientCategoryMapping.imis_patient_category_flags["male"]
        female_flag = PatientCategoryMapping.imis_patient_category_flags["female"]
        extension = Extension.construct()
        extension.url = "Gender"
        administrative_system = "http://hl7.org/fhir/administrative-gender"
        extension.valueUsageContext = UsageContext.construct()
        extension.valueUsageContext.valueCodeableConcept = CodeableConcept.construct()
        if imis_medication.patient_category & male_flag:
            coding_male = cls._build_fhir_coding(code="male", display="Male", system=administrative_system)
            cls._append_to_list_codeable_concept(extension, coding_male)
        if imis_medication.patient_category & female_flag:
            coding_female = cls._build_fhir_coding(code="female", display="Female", system=administrative_system)
            cls._append_to_list_codeable_concept(extension, coding_female)
        system_gender = "http://terminology.hl7.org/CodeSystem/usage-context-type"
        extension.valueUsageContext.code = cls._build_fhir_coding(code="gender", display="Gender", system=system_gender)
        return extension

    @classmethod
    def build_fhir_age(cls, imis_medication):
        adult_flag = PatientCategoryMapping.imis_patient_category_flags["adult"]
        child_flag = PatientCategoryMapping.imis_patient_category_flags["child"]
        extension = Extension.construct()
        extension.url = "Age"
        usage_context_system = "http://terminology.hl7.org/CodeSystem/usage-context-type"
        age_type_system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/usage-context-age-type"
        extension.valueUsageContext = UsageContext.construct()
        extension.valueUsageContext.valueCodeableConcept = CodeableConcept.construct()
        if imis_medication.patient_category & adult_flag:
            coding_adult = cls._build_fhir_coding(code="adult", display="Adult", system=age_type_system)
            cls._append_to_list_codeable_concept(extension, coding_adult)
        if imis_medication.patient_category & child_flag:
            coding_child = cls._build_fhir_coding(code="child", display="Child", system=age_type_system)
            cls._append_to_list_codeable_concept(extension, coding_child)
        extension.valueUsageContext.code = cls._build_fhir_coding(code="age", display="Age",
                                                                  system=usage_context_system)
        return extension

    @classmethod
    def _append_to_list_codeable_concept(cls, extension, coding):
        if type(extension.valueUsageContext.valueCodeableConcept.coding) is not list:
            extension.valueUsageContext.valueCodeableConcept.coding = [coding]
        else:
            extension.valueUsageContext.valueCodeableConcept.coding.append(coding)

    @classmethod
    def build_fhir_care_type(cls, imis_medication):
        if imis_medication.care_type:
            code = cls.build_fhir_act_code(imis_medication)
            display = ItemVenueTypeMapping.item_venue_type[code]
            extension = Extension.construct()
            usage_context_system = "http://terminology.hl7.org/CodeSystem/usage-context-type"
            venue_system = "http://terminology.hl7.org/CodeSystem/v3-ActCode"

            if code != "B":
                extension.url = "CareType"
                extension.valueUsageContext = UsageContext.construct()
                extension.valueUsageContext.valueCodeableConcept = CodeableConcept.construct()
                coding_venue = cls._build_fhir_coding(code=code, display=display, system=venue_system)
                cls._append_to_list_codeable_concept(extension, coding_venue)
            else:
                cls.build_fhir_both_care_type(extension, venue_system)
            extension.valueUsageContext.code = cls._build_fhir_coding(
                code="venue", display="Clinical Venue", system=usage_context_system)
            return extension

    @classmethod
    def build_fhir_both_care_type(cls, extension, venue_system):
        extension.url = "CareType"
        extension.valueUsageContext = UsageContext.construct()
        extension.valueUsageContext.valueCodeableConcept = CodeableConcept.construct()
        coding_venue = cls._build_fhir_coding(code="AMB", display="ambulatory", system=venue_system)
        cls._append_to_list_codeable_concept(extension, coding_venue)
        coding_venue = cls._build_fhir_coding(code="IMP", display="IMP", system=venue_system)
        cls._append_to_list_codeable_concept(extension, coding_venue)

    @classmethod
    def build_fhir_act_code(cls, imis_medication):
        code = ""
        if imis_medication.care_type == "O":
            code = "AMB"
        if imis_medication.care_type == "I":
            code = "IMP"
        if imis_medication.care_type == "B":
            code = "B"
        return code

    @classmethod
    def build_imis_serv_care_type(cls, imis_medication, fhir_medication, errors):
        serv_care_type = fhir_medication.useContext.text
        if not cls.valid_condition(serv_care_type is None,
                                   gettext('Missing activity definition `serv care type` attribute'), errors):
            imis_medication.care_type = serv_care_type

    @classmethod
    def build_imis_unit_price(cls, imis_medication, fhir_extension):
        imis_medication.price = fhir_extension.valueMoney.value

    @classmethod
    def build_imis_medication_type(cls, imis_medication, fhir_extension):
        imis_medication.type = fhir_extension.valueCodeableConcept.coding[0].code

    @classmethod
    def build_imis_medication_frequency(cls, imis_medication, fhir_extension):
        imis_medication.frequency = fhir_extension.valueTiming.repeat.period

    @classmethod
    def build_imis_item_usage_context(cls, imis_medication, fhir_extension):
        extensions = fhir_extension.extension
        for extension in extensions:
            if extension.url in ["Gender", "Age"]:
                usage_context_types = extension.valueUsageContext.valueCodeableConcept.coding
                cls._build_imis_item_patient_category(imis_medication, usage_context_types)
            elif extension.url == "CareType":
                cls._build_imis_item_care_type(imis_medication, extension.valueUsageContext)

    @classmethod
    def _build_imis_item_care_type(cls, imis_medication, usage_context):
        if len(usage_context.valueCodeableConcept.coding) == 2:
            imis_medication.care_type = "B"
        else:
            imis_care_type = ItemVenueTypeMapping.venue_fhir_imis[usage_context.valueCodeableConcept.coding[0].code]
            imis_medication.care_type = imis_care_type

    @classmethod
    def _build_imis_item_patient_category(cls, imis_medication, usage_context_types):
        if not imis_medication.patient_category:
            imis_medication.patient_category = 0
        number = 0
        for usage_context_type in usage_context_types:
            item_pat_cat = usage_context_type.code
            number = number | PatientCategoryMapping.imis_patient_category_flags[item_pat_cat]
        imis_medication.patient_category += number

    @classmethod
    def _build_fhir_coding(cls, code, display, system):
        coding = Coding.construct()
        coding.code = code
        coding.display = _(display)
        if GeneralConfiguration.show_system():
            coding.system = system
        return coding

    @classmethod
    def _validate_fhir_medication_identifier_code(cls, fhir_medication_identifier_code):
        if not fhir_medication_identifier_code:
            raise FHIRException(
                _('Medication FHIR without code - this field is obligatory')
            )

    @classmethod
    def build_fhir_level(cls, fhir_medication: FHIRMedication, imis_medication: Item):
        coding = cls.build_fhir_mapped_coding(ItemContextlevel.item_context_level_coding)
        extension = Extension(
            url=f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/medication-level",
            valueCodeableConcept=CodeableConcept(
                coding=[coding],
                text=coding.display
            )
        )

        if isinstance(fhir_medication.extension, list):
            fhir_medication.extension.append(extension)
        else:
            fhir_medication.extension = [extension]





import typing

SUPPORTED_FORMATS = ['json']

# fix of 'issue' type from https://github.com/nazrulworld/fhir.resources/blob/main/fhir/resources/operationoutcome.py#L31
# by overriding 'issue' property. Without this fix - there is no 'issue' field in 'OperationOutcome' model.
from fhir.resources import fhirtypes
from pydantic import Field
from fhir.resources.operationoutcome import OperationOutcome
from fhir.resources.usagecontext import UsageContext
from fhir.resources.coverage import Coverage, CoverageClass
from fhir.resources.contract import ContractSigner
from fhir.resources.claim import Claim, ClaimInsurance
from fhir.resources.claimresponse import ClaimResponse
from fhir.resources.coverageeligibilityrequest import CoverageEligibilityRequest


class OperationOutcomeV2(OperationOutcome):
    issue: typing.List[fhirtypes.OperationOutcomeIssueType] = Field(
        None,
        alias="issue",
        title="A single issue associated with the action",
        description=(
            "An error, warning, or information message that results from a system "
            "action."
        ),
        # if property is element of this resource.
        element_property=True,
    )


class UsageContextV2(UsageContext):
    code: fhirtypes.CodingType = Field(
        None,
        alias="code",
        title="Type of context being specified",
        description=(
            "A code that identifies the type of context being specified by this "
            "usage context."
        ),
        # if property is element of this resource.
        element_property=True,
    )


class CoverageV2(Coverage):
    payor: typing.List[fhirtypes.ReferenceType] = Field(
        None,
        alias="payor",
        title="Issuer of the policy",
        description=(
            "The program or plan underwriter or payor including both insurance and "
            "non-insurance agreements, such as patient-pay agreements."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["Organization", "Patient", "RelatedPerson"],
    )


class CoverageClassV2(CoverageClass):
    type: fhirtypes.CodeableConceptType = Field(
        None,
        alias="type",
        title="Type of class such as 'group' or 'plan'",
        description=(
            "The type of classification for which an insurer-specific class label "
            "or number and optional name is provided, for example may be used to "
            "identify a class of coverage or employer group, Policy, Plan."
        ),
        # if property is element of this resource.
        element_property=True,
    )


class ContractSignerV2(ContractSigner):
    party: fhirtypes.ReferenceType = Field(
        None,
        alias="party",
        title="Contract Signatory Party",
        description="Party which is a signator to this Contract.",
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=[
            "Organization",
            "Patient",
            "Practitioner",
            "PractitionerRole",
            "RelatedPerson",
        ],
    )

    type: fhirtypes.CodingType = Field(
        None,
        alias="type",
        title="Contract Signatory Role",
        description="Role of this Contract signer, e.g. notary, grantee.",
        # if property is element of this resource.
        element_property=True,
    )


class ClaimV2(Claim):
    insurance: typing.List[fhirtypes.ClaimInsuranceType] = Field(
        None,
        alias="insurance",
        title="Patient insurance information",
        description=(
            "Financial instruments for reimbursement for the health care products "
            "and services specified on the claim."
        ),
        # if property is element of this resource.
        element_property=True,
    )

    patient: fhirtypes.ReferenceType = Field(
        None,
        alias="patient",
        title="The recipient of the products and services",
        description=(
            "The party to whom the professional services and/or products have been "
            "supplied or are being considered and for whom actual or forecast "
            "reimbursement is sought."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["Patient"],
    )

    priority: fhirtypes.CodeableConceptType = Field(
        None,
        alias="priority",
        title="Desired processing ugency",
        description=(
            "The provider-required urgency of processing the request. Typical "
            "values include: stat, routine deferred."
        ),
        # if property is element of this resource.
        element_property=True,
    )

    provider: fhirtypes.ReferenceType = Field(
        None,
        alias="provider",
        title="Party responsible for the claim",
        description=(
            "The provider which is responsible for the claim, predetermination or "
            "preauthorization."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["Practitioner", "PractitionerRole", "Organization"],
    )

    type: fhirtypes.CodeableConceptType = Field(
        None,
        alias="type",
        title="Category or discipline",
        description=(
            "The category of claim, e.g. oral, pharmacy, vision, institutional, "
            "professional."
        ),
        # if property is element of this resource.
        element_property=True,
    )


class ClaimInsuranceV2(ClaimInsurance):
    coverage: fhirtypes.ReferenceType = Field(
        None,
        alias="coverage",
        title="Insurance information",
        description=(
            "Reference to the insurance card level information contained in the "
            "Coverage resource. The coverage issuing insurer will use these details"
            " to locate the patient's actual coverage within the insurer's "
            "information system."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["Coverage"],
    )


class ClaimResponseV2(ClaimResponse):
    insurer: fhirtypes.ReferenceType = Field(
        None,
        alias="insurer",
        title="Party responsible for reimbursement",
        description=(
            "The party responsible for authorization, adjudication and "
            "reimbursement."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["Organization"],
    )

    patient: fhirtypes.ReferenceType = Field(
        None,
        alias="patient",
        title="The recipient of the products and services",
        description=(
            "The party to whom the professional services and/or products have been "
            "supplied or are being considered and for whom actual for facast "
            "reimbursement is sought."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["Patient"],
    )

    type: fhirtypes.CodeableConceptType = Field(
        None,
        alias="type",
        title="More granular claim type",
        description=(
            "A finer grained suite of claim type codes which may convey additional "
            "information such as Inpatient vs Outpatient and/or a specialty "
            "service."
        ),
        # if property is element of this resource.
        element_property=True,
    )


class CoverageEligibilityRequestV2(CoverageEligibilityRequest):
    insurer: fhirtypes.ReferenceType = Field(
        None,
        alias="insurer",
        title="Coverage issuer",
        description=(
            "The Insurer who issued the coverage in question and is the recipient "
            "of the request."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["Organization"],
    )

    patient: fhirtypes.ReferenceType = Field(
        None,
        alias="patient",
        title="Intended recipient of products and services",
        description=(
            "The party who is the beneficiary of the supplied coverage and for whom"
            " eligibility is sought."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["Patient"],
    )

    type: fhirtypes.CodeableConceptType = Field(
        None,
        alias="type",
        title="More granular claim type",
        description=(
            "A finer grained suite of claim type codes which may convey additional "
            "information such as Inpatient vs Outpatient and/or a specialty "
            "service."
        ),
        # if property is element of this resource.
        element_property=True,
    )


from claim.models import Feedback, ClaimItem, ClaimService, Claim, ClaimAdmin
from django.db.models import Subquery
from medical.models import Item, Service
import core

from api_fhir_r4.configurations import GeneralConfiguration, R4ClaimConfig
from api_fhir_r4.converters import BaseFHIRConverter, CommunicationRequestConverter, ReferenceConverterMixin
from api_fhir_r4.converters.claimConverter import ClaimConverter
from api_fhir_r4.converters.patientConverter import PatientConverter
from api_fhir_r4.converters.claimAdminPractitionerConverter import ClaimAdminPractitionerConverter
from api_fhir_r4.converters.medicationConverter import MedicationConverter
from api_fhir_r4.exceptions import FHIRRequestProcessException
from api_fhir_r4.mapping.claimMapping import ClaimResponseMapping
from api_fhir_r4.models import ClaimResponseV2 as ClaimResponse
from fhir.resources.money import Money
from fhir.resources.claimresponse import ClaimResponseItem, ClaimResponseItemAdjudication, \
    ClaimResponseProcessNote, ClaimResponseTotal
from fhir.resources.reference import Reference
from fhir.resources.extension import Extension
from api_fhir_r4.utils import TimeUtils, FhirUtils


class ClaimResponseConverter(BaseFHIRConverter):

    @classmethod
    def to_fhir_obj(cls, imis_claim, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_claim_response = {}
        fhir_claim_response["created"] = TimeUtils.date().isoformat()
        cls.build_fhir_status(fhir_claim_response, imis_claim)
        cls.build_fhir_outcome(fhir_claim_response, imis_claim)
        cls.build_fhir_use(fhir_claim_response)
        fhir_claim_response = ClaimResponse(**fhir_claim_response)
        cls.build_fhir_pk(fhir_claim_response, imis_claim, reference_type)
        ClaimConverter.build_fhir_identifiers(fhir_claim_response, imis_claim)
        cls.build_fhir_items(fhir_claim_response, imis_claim, reference_type)
        cls.build_patient_reference(fhir_claim_response, imis_claim, reference_type)
        cls.build_fhir_total_list(fhir_claim_response, imis_claim)
        cls.build_fhir_communication_request_reference(fhir_claim_response, imis_claim, reference_type)
        cls.build_fhir_type(fhir_claim_response, imis_claim)
        cls.build_fhir_insurer(fhir_claim_response)
        cls.build_fhir_requestor(fhir_claim_response, imis_claim, reference_type)
        cls.build_fhir_request(fhir_claim_response, imis_claim, reference_type)
        return fhir_claim_response
               
    @classmethod
    def to_imis_obj(cls, fhir_claim_response, audit_user_id):
        errors = []
        fhir_claim_response = ClaimResponse(**fhir_claim_response)
        imis_claim = cls.get_imis_claim_from_response(fhir_claim_response)
        cls.build_imis_outcome(imis_claim, fhir_claim_response)
        cls.build_imis_items(imis_claim, fhir_claim_response)
        cls.build_imis_communication_request_reference(imis_claim, fhir_claim_response)
        cls.build_imis_type(imis_claim, fhir_claim_response)
        cls.build_imis_status(imis_claim, fhir_claim_response)
        cls.build_imis_requestor(imis_claim, fhir_claim_response)
        return imis_claim

    @classmethod
    def get_reference_obj_uuid(cls, imis_claim):
        return imis_claim.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_claim):
        return imis_claim.id

    @classmethod
    def get_reference_obj_code(cls, imis_claim):
        return imis_claim.code

    @classmethod
    def build_fhir_outcome(cls, fhir_claim_response, imis_claim):
        status = imis_claim.status
        outcome = ClaimResponseMapping.claim_outcome[f'{status}']
        fhir_claim_response["outcome"] = outcome

    @classmethod
    def build_imis_outcome(cls, imis_claim, fhir_claim_response):
        if fhir_claim_response.outcome is not None:
            status_code = cls.get_status_code_by_display(fhir_claim_response.outcome)
            imis_claim.status = status_code

    @classmethod
    def get_imis_claim_from_response(cls, fhir_claim_response):
        claim_uuid = fhir_claim_response.id
        try:
            return Claim.objects.get(uuid=claim_uuid)
        except Claim.DoesNotExit:
            raise FHIRRequestProcessException(F"Claim Response cannot be created from scratch, "
                                              f"IMIS instance for reference {claim_uuid} was not found.")

    @classmethod
    def get_status_code_by_display(cls, claim_response_display):
        for code, display in ClaimResponseMapping.claim_outcome.items():
            if display == claim_response_display:
                return code
        return None

    @classmethod
    def get_imis_claim_feedback(cls, imis_claim):
        try:
            feedback = imis_claim.feedback
        except Feedback.DoesNotExist:
            feedback = None
        return feedback

    @classmethod
    def build_patient_reference(cls, fhir_claim_response, imis_claim, reference_type):
        fhir_claim_response.patient = PatientConverter\
            .build_fhir_resource_reference(imis_claim.insuree, reference_type=reference_type)

    @classmethod
    def build_fhir_total_list(cls, fhir_claim_response, imis_claim):
        total_claimed = 0
        total_approved = 0
        total_adjusted = 0
        total_rejected = 0
        total_valuated = 0
        if fhir_claim_response:
            items = fhir_claim_response.item
            for item in items:
                for adjudication in item.adjudication:
                    rejection_reason = adjudication.reason.coding[0].code
                    if rejection_reason == '0':
                        if adjudication.category.coding[0].code == str(Claim.STATUS_ENTERED):
                            total_claimed += adjudication.amount.value*adjudication.value
                        if adjudication.category.coding[0].code == str(Claim.STATUS_PROCESSED):
                            total_adjusted += adjudication.amount.value*adjudication.value
                        if adjudication.category.coding[0].code == str(Claim.STATUS_CHECKED):
                            total_approved += adjudication.amount.value*adjudication.value
                        if adjudication.category.coding[0].code == str(Claim.STATUS_VALUATED):
                            total_valuated += adjudication.amount.value*adjudication.value
                    else:
                        total_claimed += adjudication.amount.value*adjudication.value

        fhir_total = []
        if imis_claim.status >= Claim.STATUS_ENTERED:
            fhir_total.append(cls.build_fhir_total(imis_claim, Claim.STATUS_ENTERED, total_claimed))
        if imis_claim.status >= Claim.STATUS_CHECKED:
            fhir_total.append(cls.build_fhir_total(imis_claim, Claim.STATUS_CHECKED, total_approved))
        if imis_claim.status >= Claim.STATUS_PROCESSED:
            fhir_total.append(cls.build_fhir_total(imis_claim, Claim.STATUS_PROCESSED, total_adjusted))
        if imis_claim.status == Claim.STATUS_VALUATED:
            fhir_total.append(cls.build_fhir_total(imis_claim, Claim.STATUS_VALUATED, total_valuated))

        if len(fhir_total) > 0:
            fhir_claim_response.total = fhir_total

    @classmethod
    def build_fhir_total(cls, imis_claim, claim_status, total):
        fhir_total = ClaimResponseTotal.construct()
        money = Money.construct()
        fhir_total.amount = money
        fhir_total.category = cls.build_codeable_concept(
            system=ClaimResponseMapping.claim_status_system,
            code=claim_status,
            display=ClaimResponseMapping.claim_status[f'{claim_status}']
        )
        fhir_total.amount.value = total
        if hasattr(core, 'currency'):
            fhir_total.amount.currency = core.currency
        return fhir_total

    @classmethod
    def build_fhir_communication_request_reference(cls, fhir_claim_response, imis_claim, reference_type):
        try:
            if imis_claim.feedback is not None:
                request = CommunicationRequestConverter\
                    .build_fhir_resource_reference(imis_claim.feedback, reference_type=reference_type)
                fhir_claim_response.communicationRequest = [request]
        except Feedback.DoesNotExist:
            pass

    @classmethod
    def build_imis_communication_request_reference(cls, imis_claim, fhir_claim_response):
        try:
            if fhir_claim_response.communicationRequest:
                request = fhir_claim_response.communicationRequest[0]
                _, feedback_id = request.reference.split("/")
                imis_claim.feedback = Feedback.objects.get(uuid=feedback_id)
        except Feedback.DoesNotExist:
            pass

    @classmethod
    def build_fhir_type(cls, fhir_claim_response, imis_claim):
        if imis_claim.visit_type:
            fhir_claim_response.type = cls.build_codeable_concept(
                system=ClaimResponseMapping.visit_type_system,
                code=imis_claim.visit_type,
                display=ClaimResponseMapping.visit_type[f'{imis_claim.visit_type}']
            )

    @classmethod
    def build_imis_type(cls, imis_claim, fhir_claim_response):
        if fhir_claim_response.type:
            coding = fhir_claim_response.type.coding
            if coding and len(coding) > 0:
                visit_type = fhir_claim_response.type.coding[0].code
                imis_claim.visit_type = visit_type

    _REVIEW_STATUS_DISPLAY = {
        1: "Idle",
        2: "Not Selected",
        4: "Selected for Review",
        8: "Reviewed",
        16: "ByPassed"
    }

    @classmethod
    def build_fhir_status(cls, fhir_claim_response, imis_claim):
        fhir_claim_response["status"] = "active"

    @classmethod
    def build_imis_status(cls, imis_claim, fhir_claim_response):
        fhir_status_display = fhir_claim_response.status
        for status_code, status_display in cls._REVIEW_STATUS_DISPLAY.items():
            if fhir_status_display == status_display:
                imis_claim.review_status = status_code
                break

    @classmethod
    def build_fhir_use(cls, fhir_claim_response):
        fhir_claim_response["use"] = "claim"

    @classmethod
    def build_fhir_insurer(cls, fhir_claim_response):
        fhir_claim_response.insurer = Reference.construct()
        fhir_claim_response.insurer.reference = "openIMIS"

    @classmethod
    def build_fhir_items(cls, fhir_claim_response, imis_claim, reference_type):
        fhir_claim_response.item = []
        cls.build_fhir_items_for_imis_items(fhir_claim_response, imis_claim, reference_type)
        cls.build_fhir_items_for_imis_services(fhir_claim_response, imis_claim, reference_type)

    @classmethod
    def build_fhir_items_for_imis_services(cls, fhir_claim_response, imis_claim, reference_type):
        for claim_service in imis_claim.services.filter(validity_to=None).all():
            if claim_service:
                item_type = R4ClaimConfig.get_fhir_claim_service_code()
                cls.build_fhir_item(fhir_claim_response, claim_service, item_type, claim_service.rejection_reason, imis_claim, reference_type)

    @classmethod
    def build_fhir_items_for_imis_items(cls, fhir_claim_response, imis_claim, reference_type):
        for claim_item in imis_claim.items.filter(validity_to=None).all():
            if claim_item:
                item_type = R4ClaimConfig.get_fhir_claim_item_code()
                cls.build_fhir_item(fhir_claim_response, claim_item, item_type, claim_item.rejection_reason, imis_claim, reference_type)


    @classmethod
    def build_imis_items(cls, imis_claim: Claim, fhir_claim_response: ClaimResponse):
        # Added new attributes since items shouldn't be saved during mapping to imis
        imis_claim.claim_items = []
        imis_claim.claim_services = []
        for item in fhir_claim_response.item:
            cls._build_imis_claim_item(imis_claim, fhir_claim_response, item)  # same for item and service

    @classmethod
    def _build_response_items(cls, fhir_claim_response, claim_item, imis_service,
                              type, rejected_reason, imis_claim, reference_type):
        cls.build_fhir_item(fhir_claim_response, claim_item, imis_service,
                            type, rejected_reason, imis_claim, reference_type)

    @classmethod
    def get_imis_claim_item_by_code(cls, code, imis_claim_id):
        item_code_qs = Item.objects.filter(code=code)
        result = ClaimItem.objects.filter(item_id__in=Subquery(item_code_qs.values('id')), claim_id=imis_claim_id)
        return result[0] if len(result) > 0 else None

    @classmethod
    def _build_imis_claim_item(cls, imis_claim, fhir_claim_response: ClaimResponse, item: ClaimResponseItem):
        extension = item.extension[0]
        _, resource_id = extension.valueReference.reference.split("/")

        if extension.valueReference.type == 'Medication':
            imis_item = Item.objects.get(uuid=resource_id)
            claim_item = ClaimItem.objects.get(claim=imis_claim, item=imis_item)
        elif extension.valueReference.type == 'ActivityDefinition':
            imis_service = Service.objects.get(uuid=resource_id)
            claim_item = ClaimService.objects.get(claim=imis_claim, service=imis_service)
        else:
            raise FHIRRequestProcessException(F"Unknnown serviced item type: {extension.url}")

        for next_adjudication in item.adjudication:
            cls.adjudication_to_item(next_adjudication, claim_item, fhir_claim_response)

        if isinstance(claim_item, ClaimItem):
            if imis_claim.claim_items is not list:
                imis_claim.claim_items = [claim_item]
            else:
                imis_claim.claim_items.append(claim_item)
        elif isinstance(claim_item, ClaimService):
            if imis_claim.claim_services is not list:
                imis_claim.claim_services = [claim_item]
            else:
                imis_claim.claim_services.append(claim_item)

    @classmethod
    def _build_imis_claim_service(cls, item: ClaimItem, imis_claim):
        pass

    @classmethod
    def get_imis_claim_service_by_code(cls, code, imis_claim_id):
        service_code_qs = Service.objects.filter(code=code)
        result = ClaimService.objects.filter(service_id__in=Subquery(service_code_qs.values('id')),
                                             claim_id=imis_claim_id)
        return result[0] if len(result) > 0 else None

    @classmethod
    def build_fhir_item(cls, fhir_claim_response, item, type, rejected_reason, imis_claim, reference_type):
        claim_response_item = ClaimResponseItem.construct()
        claim_response_item.itemSequence = FhirUtils.get_next_array_sequential_id(fhir_claim_response.item)

        adjudication = cls.build_fhir_item_adjudication(item, rejected_reason, imis_claim)
        claim_response_item.adjudication = adjudication

        if type == "item":
            service_type = "Medication"
            serviced_item = item.item
        elif type == "service":
            service_type = "ActivityDefinition"
            serviced_item = item.service
        else:
            raise FHIRRequestProcessException(F"Unknown type of serviced product: {type}")

        serviced_extension = cls.build_serviced_extension(serviced_item, service_type, reference_type)

        if claim_response_item.extension is not list:
            claim_response_item.extension = [serviced_extension]
        else:
            claim_response_item.extension.append(serviced_extension)

        note = cls.build_process_note(fhir_claim_response, item.price_origin)
        if note:
            claim_response_item.noteNumber = [note.number]

        fhir_claim_response.item.append(claim_response_item)

    @classmethod
    def build_serviced_extension(cls, serviced, service_type, reference_type):
        reference = Reference.construct()
        extension = Extension.construct()
        extension.valueReference = reference
        extension.url = f'{GeneralConfiguration.get_system_base_url()}StructureDefinition/claim-item-reference'
        extension.valueReference = MedicationConverter\
            .build_fhir_resource_reference(serviced, service_type, reference_type=reference_type, display=serviced.code)
        return extension

    @classmethod
    def __build_item_price(cls, item_price):
        price = Money.construct()
        if hasattr(core, 'currency'):
            price.currency = core.currency
        price.value = item_price
        return price

    @classmethod
    def __build_adjudication(cls, item, rejected_reason, amount, category, quantity, explicit_amount=False):
        adjudication = ClaimResponseItemAdjudication.construct()
        adjudication.reason = cls.build_fhir_adjudication_reason(item, rejected_reason)
        adjudication.amount = amount
        adjudication.category = category
        adjudication.value = quantity
        return adjudication

    @classmethod
    def build_fhir_item_adjudication(cls, item, rejected_reason, imis_claim):
        def build_asked_adjudication(status, price):
            category = cls.build_codeable_concept(
                system=ClaimResponseMapping.claim_status_system,
                code=status,
                display=ClaimResponseMapping.claim_status[f'{status}']
            )
            adjudication = cls.__build_adjudication(item, rejected_reason, price, category, item.qty_provided, True)
            return adjudication

        def build_processed_adjudication(status, price):
            category = cls.build_codeable_concept(
                system=ClaimResponseMapping.claim_status_system,
                code=status,
                display=ClaimResponseMapping.claim_status[f'{status}']
            )
            if item.qty_approved is not None and item.qty_approved != 0.0:
                quantity = item.qty_approved
            else:
                quantity = item.qty_provided
            adjudication = cls.__build_adjudication(item, rejected_reason, price, category, quantity)
            return adjudication

        price_asked = cls.__build_item_price(item.price_asked)
        adjudications = []

        if rejected_reason == 0 and imis_claim.status != 1:
            if imis_claim.status >= Claim.STATUS_ENTERED:
                adjudications.append(build_asked_adjudication(Claim.STATUS_ENTERED, price_asked))

            if imis_claim.status >= Claim.STATUS_CHECKED:
                if item.price_approved:
                    price_approved = cls.__build_item_price(item.price_approved)
                    adjudications.append(build_processed_adjudication(Claim.STATUS_CHECKED, price_approved))
                else:
                    adjudications.append(build_processed_adjudication(Claim.STATUS_CHECKED, price_asked))
            if imis_claim.status >= Claim.STATUS_PROCESSED:
                if item.price_adjusted:
                    price_adjusted = cls.__build_item_price(item.price_adjusted)
                    adjudications.append(build_processed_adjudication(Claim.STATUS_PROCESSED, price_adjusted))
                else:
                    adjudications.append(build_processed_adjudication(Claim.STATUS_PROCESSED, price_asked))
            if imis_claim.status == Claim.STATUS_VALUATED:
                if item.price_valuated:
                    price_valuated = cls.__build_item_price(item.price_valuated)
                    adjudications.append(build_processed_adjudication(Claim.STATUS_VALUATED, price_valuated))
                else:
                    adjudications.append(build_processed_adjudication(Claim.STATUS_VALUATED, price_asked))
        else:
            adjudications.append(build_asked_adjudication(1, price_asked))

        return adjudications

    @classmethod
    def build_fhir_adjudication_reason(cls, item, rejected_reason):
        code = "0" if not rejected_reason else rejected_reason
        return cls.build_codeable_concept(
            system=ClaimResponseMapping.rejection_reason_system,
            code=code,
            display=ClaimResponseMapping.rejection_reason[int(code)]
        )

    @classmethod
    def adjudication_to_item(cls, adjudication, claim_item, fhir_claim_response):
        status = int(adjudication.category.coding[0].code)
        if status == 1:
            cls.build_item_rejection(claim_item, adjudication)
        if status == 2:
            cls.build_item_entered(claim_item, adjudication)
        if status == 4:
            cls.build_item_checked(claim_item, adjudication)
        if status == 8:
            cls.build_item_processed(claim_item, adjudication)
        if status == 16:
            cls.build_item_valuated(claim_item, adjudication)
        claim_item.status = status
        return claim_item

    @classmethod
    def build_item_rejection(cls, claim_item, adjudication):
        claim_item.rejection_reason = int(adjudication.reason.coding[0].code)
        cls.build_item_entered(claim_item, adjudication)

    @classmethod
    def build_item_entered(cls, claim_item, adjudication):
        claim_item.qty_provided = adjudication.value
        claim_item.price_asked = adjudication.amount.value

    @classmethod
    def build_item_checked(cls, claim_item, adjudication):
        if adjudication.value and adjudication.value != claim_item.qty_provided:
            claim_item.qty_approved = adjudication.value
        if adjudication.amount and adjudication.amount.value != claim_item.price_asked:
            claim_item.price_approved = adjudication.amount.value

    @classmethod
    def build_item_processed(cls, claim_item, adjudication):
        if adjudication.value and adjudication.value != claim_item.qty_provided:
            claim_item.qty_approved = adjudication.value
        if adjudication.amount and adjudication.amount.value != claim_item.price_asked:
            claim_item.price_adjusted = adjudication.amount.value

    @classmethod
    def build_item_valuated(cls, claim_item, adjudication):
        if adjudication.value and adjudication.value != claim_item.qty_provided:
            claim_item.qty_approved = adjudication.value
        if adjudication.amount and adjudication.amount.value != claim_item.price_asked * claim_item.qty_provided:
            claim_item.price_valuated = adjudication.amount.value

    @classmethod
    def build_process_note(cls, fhir_claim_response, string_value):
        result = None
        if string_value:
            note = ClaimResponseProcessNote.construct()
            note.text = string_value
            note.number = FhirUtils.get_next_array_sequential_id(fhir_claim_response.processNote)
            if fhir_claim_response.processNote is not list:
                fhir_claim_response.processNote = [note]
            else:
                fhir_claim_response.processNote.append(note)
            result = note
        return result

    @classmethod
    def build_fhir_requestor(cls, fhir_claim_response, imis_claim, reference_type):
        if imis_claim.admin is not None:
            fhir_claim_response.requestor = ClaimAdminPractitionerConverter\
                .build_fhir_resource_reference(imis_claim.admin, reference_type=reference_type)

    @classmethod
    def build_imis_requestor(cls, imis_claim, fhir_claim_response):
        if fhir_claim_response.requestor is not None:
            requestor = fhir_claim_response.requestor
            _, claim_admin_uuid = requestor.reference.split("/")
            imis_claim.admin = ClaimAdmin.objects.get(uuid=claim_admin_uuid)

    @classmethod
    def build_fhir_request(cls, fhir_claim_response: ClaimResponse, imis_claim: Claim, reference_type):
        fhir_claim_response.request = ClaimConverter\
            .build_fhir_resource_reference(imis_claim, reference_type=reference_type)



import datetime

from django.db.models import Prefetch
from rest_framework import mixins
from rest_framework.viewsets import GenericViewSet

from api_fhir_r4.permissions import FHIRApiCoverageRequestPermissions
from api_fhir_r4.serializers import ContractSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from insuree.models import InsureePolicy
from policy.models import Policy


class ContractViewSet(BaseFHIRView, mixins.RetrieveModelMixin, mixins.ListModelMixin, mixins.CreateModelMixin,
                      GenericViewSet):
    lookup_field = 'uuid'
    serializer_class = ContractSerializer
    permission_classes = (FHIRApiCoverageRequestPermissions,)

    def list(self, request, *args, **kwargs):
        queryset = self.get_queryset() \
            .select_related('product') \
            .select_related('officer') \
            .select_related('family__head_insuree') \
            .select_related('family__location') \
            .prefetch_related(
            Prefetch('insuree_policies',
                     queryset=InsureePolicy.objects.filter(validity_to__isnull=True).select_related('insuree'))
        )
        refDate = request.GET.get('refDate')
        refEndDate = request.GET.get('refEndDate')
        identifier = request.GET.get("identifier")
        if identifier:
            queryset = queryset.filter(chf_id=identifier)
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by('validity_from')
            if refDate != None:
                isValidDate = True
                try:
                    datevar = datetime.datetime.strptime(refDate, "%Y-%m-%d").date()
                except ValueError:
                    isValidDate = False
                queryset = queryset.filter(validity_from__gte=datevar)
            if refEndDate != None:
                isValidDate = True
                try:
                    datevar = datetime.datetime.strptime(refEndDate, "%Y-%m-%d").date()
                except ValueError:
                    isValidDate = False
                queryset = queryset.filter(validity_from__lt=datevar)

        serializer = ContractSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def get_queryset(self):
        queryset = Policy.get_queryset(None, self.request.user)
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(queryset)



import copy

from location.models import Location

from api_fhir_r4.converters import LocationConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class LocationSerializer(BaseFHIRSerializer):
    fhirConverter = LocationConverter()

    def create(self, validated_data):
        copied_data = copy.deepcopy(validated_data)
        del copied_data['_state']
        return Location.objects.create(**copied_data)

    def update(self, instance, validated_data):
        instance.code = validated_data.get('code', instance.code)
        instance.name = validated_data.get('name', instance.name)
        instance.type = validated_data.get('type', instance.type)
        instance.parent_id = validated_data.get('parent_id', instance.parent_id)
        instance.audit_user_id = self.get_audit_user_id()
        instance.save()
        return instance



from api_fhir_r4.utils.functionUtils import FunctionUtils
from api_fhir_r4.utils.timeUtils import TimeUtils
from api_fhir_r4.utils.fhirUtils import FhirUtils
from api_fhir_r4.utils.dbManagerUtils import DbManagerUtils
